"""federated_content_connector signals."""
from logging import getLogger

from federated_content_connector.models import CourseDetails
from federated_content_connector.tasks import import_course_metadata

LOGGER = getLogger(__name__)


def handle_courseoverview_import_course_details(sender, courserun_key, **kwargs):  # pylint: disable=unused-argument
    """Handle CourseOverview.import_course_metadata signal."""
    LOGGER.info(
        f"[FEDERATED_CONTENT_CONNECTOR] CourseOverview.import_course_metadata signal received. Key: [{courserun_key}]"
    )

    courserun_keys = [courserun_key]
    import_course_metadata.delay(courserun_keys)


def handle_courseoverview_delete_course_details(sender, courserun_key, **kwargs):  # pylint: disable=unused-argument
    """Handle CourseOverview.delete_course_metadata signal."""
    LOGGER.info(
        f"[FEDERATED_CONTENT_CONNECTOR] CourseOverview.delete_course_metadata signal received. Key: [{courserun_key}]"
    )
    CourseDetails.objects.filter(id=courserun_key).delete()



# Generated by Django 3.2.19 on 2023-06-07 11:07

from django.db import migrations, models
import django_extensions.db.fields
import opaque_keys.edx.django.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CourseDetails',
            fields=[
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('id', opaque_keys.edx.django.models.CourseKeyField(db_index=True, help_text='Courserun key', max_length=255, primary_key=True, serialize=False)),
                ('course_type', models.CharField(help_text='Type of course. For example Masters, Verified, Audit, executive-education-2u,  etc', max_length=255)),
                ('product_source', models.CharField(help_text='Tells about the origin of a course. For example, edx, 2u', max_length=255)),
                ('start_date', models.DateTimeField(help_text='The start date of the course.', null=True)),
                ('end_date', models.DateTimeField(help_text='The end date of the course.', null=True)),
                ('enroll_by', models.DateTimeField(help_text='The suggested deadline for enrollment.', null=True)),
            ],
        ),
    ]



"""
Tests for the `federated_content_connector` signals module.
"""

from unittest import TestCase
from unittest.mock import MagicMock, patch

import pytest
from opaque_keys.edx.keys import CourseKey

from federated_content_connector.management.commands.import_course_runs_metadata import CourseMetadataImporter
from federated_content_connector.management.commands.tests.test_utils import side_effect_func
from federated_content_connector.models import CourseDetails
from federated_content_connector.signals import (
    handle_courseoverview_delete_course_details,
    handle_courseoverview_import_course_details,
)


@pytest.mark.django_db
class TestSignals(TestCase):
    """
    Tests class for signals.
    """

    def setUp(self):
        super().setUp()

        self.courserun_locators = [
            CourseKey.from_string("course-v1:edX+DemoX+Demo_Course"),
            CourseKey.from_string("course-v1:edX+E2E-101+course"),
        ]

    @patch.object(CourseMetadataImporter, 'get_api_client')
    @patch.object(CourseMetadataImporter, 'courserun_locators_to_import')
    def test_handle_courseoverview_import_course_details(self, mocked_courserun_locator, mocked_get_api_client):
        """
        Verify that `handle_courseoverview_import_course_details` signal handler work as expected.
        """
        # Mock api client and api responses
        mocked_get_api_client.return_value = MagicMock()
        mocked_get_api_client.return_value.get = MagicMock(side_effect=side_effect_func)
        mocked_courserun_locator.return_value = [self.courserun_locators[1]]

        assert CourseDetails.objects.count() == 0

        handle_courseoverview_import_course_details(None, str(self.courserun_locators[1]))

        assert CourseDetails.objects.count() == 1
        assert CourseDetails.objects.filter(id=self.courserun_locators[0]).exists() is False
        assert CourseDetails.objects.filter(id=self.courserun_locators[1]).exists()

    @patch.object(CourseMetadataImporter, 'get_api_client')
    @patch.object(CourseMetadataImporter, 'courserun_locators_to_import')
    def test_handle_courseoverview_delete_course_details(self, mocked_courserun_locator, mocked_get_api_client):
        """
        Verify that `handle_courseoverview_delete_course_details` signal handler work as expected.
        """
        # Mock api client and api responses
        mocked_get_api_client.return_value = MagicMock()
        mocked_get_api_client.return_value.get = MagicMock(side_effect=side_effect_func)
        mocked_courserun_locator.return_value = self.courserun_locators

        # Import data
        CourseMetadataImporter.import_all_courses_metadata()

        assert CourseDetails.objects.count() == 2

        handle_courseoverview_delete_course_details(None, self.courserun_locators[0])

        assert CourseDetails.objects.count() == 1
        assert CourseDetails.objects.filter(id=self.courserun_locators[0]).exists() is False
        assert CourseDetails.objects.filter(id=self.courserun_locators[1]).exists()



"""
federated_content_connector Django application initialization.
"""

from django.apps import AppConfig
from edx_django_utils.plugins.constants import PluginSettings, PluginSignals


class FederatedContentConnectorConfig(AppConfig):
    """
    Configuration for the federated_content_connector Django application.
    """

    name = 'federated_content_connector'
    label = 'federated_content_connector'
    verbose_name = "Federated Content Connector"
    plugin_app = {
        PluginSettings.CONFIG: {
            'lms.djangoapp': {
                'common': {
                    PluginSettings.RELATIVE_PATH: 'settings.common',
                },
                'production': {
                    PluginSettings.RELATIVE_PATH: 'settings.production',
                },
            }
        },
        PluginSignals.CONFIG: {
            'lms.djangoapp': {
                PluginSignals.RECEIVERS: [
                    {
                        PluginSignals.SIGNAL_PATH: 'openedx.core.djangoapps.content.course_overviews.signals.IMPORT_COURSE_DETAILS',  # noqa: pylint: disable=line-too-long
                        PluginSignals.RECEIVER_FUNC_NAME: 'handle_courseoverview_import_course_details',
                    },
                    {
                        PluginSignals.SIGNAL_PATH: 'openedx.core.djangoapps.content.course_overviews.signals.DELETE_COURSE_DETAILS',  # noqa: pylint: disable=line-too-long
                        PluginSignals.RECEIVER_FUNC_NAME: 'handle_courseoverview_delete_course_details',
                    },
                ],
            }
        }
    }



# Generated by Django 4.2.14 on 2024-07-25 18:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('federated_content_connector', '0003_reset_refresh_timestamp'),
    ]

    operations = [
        migrations.AddField(
            model_name='coursedetails',
            name='course_key',
            field=models.CharField(blank=True, default='', help_text='The top top level course key associated with the courserun key', max_length=255),
        ),
    ]



""""Management command to refresh course metadata."""

import logging
from datetime import datetime, timedelta

from django.core.management import BaseCommand
from django.utils.timezone import now

from federated_content_connector.course_metadata_importer import CourseMetadataImporter
from federated_content_connector.models import CourseDetailsImportStatus

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """Management command to refresh course metadata"""

    help = "Refresh course metadata"

    def handle(self, *args, **options):
        self.refresh_courses_metadata()

    @classmethod
    def refresh_courses_metadata(cls):
        """Refresh courses updated after last refresh."""
        timestamp_format = CourseDetailsImportStatus.TIMESTAMP_FORMAT
        course_data_modified_timestamps = []

        last_successful_import_timestamp = CourseDetailsImportStatus.last_successful_import_timestamp()
        # This will be true only once in a lifetime
        if last_successful_import_timestamp is None:
            logger.info('[REFRESH_COURSE_METADATA] No previous timstamp found.')
            timestamp = now() - timedelta(hours=1)
            last_successful_import_timestamp = timestamp.strftime(timestamp_format)
            CourseDetailsImportStatus.save_last_successful_import_timestamp(last_successful_import_timestamp)

        logger.info(f'[REFRESH_COURSE_METADATA] Refresh Started. Timestamp: [{last_successful_import_timestamp}]')

        for courses in CourseMetadataImporter.courses(last_successful_import_timestamp):
            course_data_modified_timestamps.extend([course['data_modified_timestamp'] for course in courses])

            courserun_with_course_uuids = cls.courseruns_to_update(courses)
            courserun_keys = courserun_with_course_uuids.keys()

            logger.info(f'[REFRESH_COURSE_METADATA] Processing. Courseruns: [{courserun_keys}]')

            processed_courses_details = CourseMetadataImporter.process_courses_details(
                courses,
                courserun_with_course_uuids
            )
            CourseMetadataImporter.store_courses_details(processed_courses_details)

            logger.info(f'[REFRESH_COURSE_METADATA] Processing Completed. Courseruns: [{courserun_keys}]')

        # Sort course timestamps in descending order and store first timestamp as last_successful_import_timestamp
        if course_data_modified_timestamps:
            logger.info(f'[REFRESH_COURSE_METADATA] All Course Timestamps: [{course_data_modified_timestamps}]')
            sorted_timestamps = sorted(
                course_data_modified_timestamps,
                key=lambda timestamp: datetime.strptime(timestamp, timestamp_format),
                reverse=True
            )
            next_timestamp = sorted_timestamps[0]
            CourseDetailsImportStatus.save_last_successful_import_timestamp(next_timestamp)

            logger.info(f'[REFRESH_COURSE_METADATA] Next Timestamp: [{next_timestamp}]')

        logger.info('[REFRESH_COURSE_METADATA] Refresh Completed.')

    @classmethod
    def courseruns_to_update(cls, courses):
        """Return a map of courserun key and course uuid."""
        courserun_with_course_uuids = {}
        for course in courses:
            for courserun in course.get('course_runs', []):
                courserun_with_course_uuids[courserun.get('key')] = courserun.get('course_uuid')

        return courserun_with_course_uuids



"""Constants for the federated content connector."""

EXEC_ED_LANDING_PAGE = 'https://www.getsmarter.com/account'
EXEC_ED_COURSE_TYPE = 'executive-education-2u'
BOOTCAMP_2U = 'bootcamp-2u'
PRODUCT_SOURCE_2U = '2u'



"""Course metadata importer."""

import logging
from urllib.parse import quote_plus, urlencode, urljoin

import backoff
from common.djangoapps.course_modes.models import CourseMode
from django.contrib.auth import get_user_model
from openedx.core.djangoapps.catalog.models import CatalogIntegration
from openedx.core.djangoapps.catalog.utils import get_catalog_api_base_url, get_catalog_api_client
from openedx.core.djangoapps.content.course_overviews.models import CourseOverview

from federated_content_connector.models import CourseDetails

BEST_MODE_ORDER = [
    CourseMode.VERIFIED,
    CourseMode.PROFESSIONAL,
    CourseMode.NO_ID_PROFESSIONAL_MODE,
    CourseMode.UNPAID_EXECUTIVE_EDUCATION,
    CourseMode.AUDIT,
]

logger = logging.getLogger(__name__)
User = get_user_model()


class CourseMetadataImporter:
    """
    Import course metadata from discovery.
    """

    LOG_PREFIX = 'COURSE_METADATA_IMPORTER'

    @classmethod
    def get_api_client(cls):
        """
        Return discovery api client.
        """
        catalog_integration = CatalogIntegration.current()
        username = catalog_integration.service_username

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            logger.exception(
                f'Failed to create API client. Service user {username} does not exist.'
            )
            raise

        return get_catalog_api_client(user)

    @classmethod
    def import_all_courses_metadata(cls):
        """
        Import course metadata for all courses.
        """
        logger.info(f'[{cls.LOG_PREFIX}] Course metadata import started for all courses.')
        all_active_courserun_locators = cls.courserun_locators_to_import()
        cls.import_courses_metadata(all_active_courserun_locators)
        logger.info(f'[{cls.LOG_PREFIX}] Course metadata import completed for all courses.')

    @classmethod
    def import_specific_courses_metadata(cls, courserun_locators):
        """
        Import course metadata for specific courses.

        Args:
            courserun_locators (list): list of courserun locator objects
        """
        logger.info(f'[{cls.LOG_PREFIX}] Course metadata import started for courses. {courserun_locators}')
        cls.import_courses_metadata(courserun_locators)
        logger.info(f'[{cls.LOG_PREFIX}] Course metadata import completed for courses. {courserun_locators}')

    @classmethod
    def import_courses_metadata(cls, courserun_locators):
        """
        Import course metadata for given course locators.

        Args:
            courserun_locators (list): list of courserun locator objects
        """
        logger.info(f'[{cls.LOG_PREFIX}] Course metadata import started.')

        for courserun_locators_chunk in cls.chunks(courserun_locators):

            # convert course locator objects to courserun keys
            courserun_keys = list(map(str, courserun_locators_chunk))

            logger.info(f'[{cls.LOG_PREFIX}] Importing metadata. Courses: {courserun_keys}')

            course_details, courserun_with_course_uuids = cls.fetch_courses_details(
                courserun_locators_chunk,
                get_catalog_api_base_url()
            )
            processed_courses_details = cls.process_courses_details(course_details, courserun_with_course_uuids)
            cls.store_courses_details(processed_courses_details)

            logger.info(f'[{cls.LOG_PREFIX}] Import completed. Courses: {courserun_keys}')

    @classmethod
    def courserun_locators_to_import(cls):
        """
        Construct list of all course locators for which we want to import data.
        """
        return list(CourseOverview.objects.all().values_list('id', flat=True))

    @classmethod
    def fetch_courses_details(cls, courserun_locators, api_base_url):
        """
        Fetch the course data from discovery using `/api/v1/courses` endpoint.
        """
        courserun_with_course_uuids = cls.fetch_course_uuids(api_base_url, courserun_locators)
        course_uuids = courserun_with_course_uuids.values()
        course_uuids_str = ','.join(course_uuids)

        logger.info(f'[{cls.LOG_PREFIX}] Fetching details from discovery. Course UUIDs {course_uuids}.')
        api_url = urljoin(
            f"{api_base_url}/", f"courses/?limit=50&include_hidden_course_runs=1&uuids={course_uuids_str}"
        )
        response = cls.get_response_from_api(api_url)
        courses_details = response.json()
        results = courses_details.get('results', [])

        return results, courserun_with_course_uuids

    @classmethod
    def fetch_course_uuids(cls, api_base_url, courserun_locators):
        """
        Return a map of courserun key and course uuid.
        """
        courserun_keys = list(map(str, courserun_locators))
        encoded_courserun_keys = ','.join(map(quote_plus, courserun_keys))

        logger.info(f'[{cls.LOG_PREFIX}] Fetching uuids for Courseruns {encoded_courserun_keys}')
        api_url = urljoin(
            f"{api_base_url}/", f"course_runs/?limit=50&include_hidden_course_runs=1&keys={encoded_courserun_keys}"
        )
        response = cls.get_response_from_api(api_url)
        courses_details = response.json()
        results = courses_details.get('results', [])

        courserun_with_course_uuids = {}
        for result in results:
            courserun_key = result.get('key')

            if courserun_key not in courserun_keys:
                continue

            courserun_with_course_uuids[courserun_key] = result.get('course_uuid')

        return courserun_with_course_uuids

    @classmethod
    def courses(cls, timestamp):
        """Fetch courses updated since `timestamp`."""
        query_params = {
            'timestamp': timestamp,
            'limit': 50,
            'include_hidden_course_runs': 1,
        }

        api_base_url = get_catalog_api_base_url()
        params = urlencode(query_params)
        api_url = urljoin(f"{api_base_url}/", f"courses/?{params}")
        results, next_url, total = cls.get_api_reponse(api_url)
        logger.info(f'[{cls.LOG_PREFIX}] Total Records are {total}')
        yield results

        while next_url:
            results, next_url, __ = cls.get_api_reponse(next_url)
            yield results

    @classmethod
    def get_api_reponse(cls, api_url):
        """Get response from API."""
        response = cls.get_response_from_api(api_url)
        courses = response.json()
        results = courses.get('results', [])
        return results, courses.get('next'), courses.get('count')

    @classmethod
    @backoff.on_exception(
        backoff.expo,
        Exception,
        max_tries=3,
        logger=logger,
    )
    def get_response_from_api(cls, api_url):
        """
        Call api endpoint and return response.
        """
        logger.info(f'[{cls.LOG_PREFIX}] API Call: URL: [{api_url}]')
        client = cls.get_api_client()
        response = client.get(api_url)
        response.raise_for_status()
        return response

    @classmethod
    def process_courses_details(cls, courses_details, courserun_with_course_uuids):
        """
        Parse and extract the minimal data that we need.
        """
        courses = {}
        for courserun_key, course_uuid in courserun_with_course_uuids.items():
            logger.info(f'[{cls.LOG_PREFIX}] Process. CourserunKey: {courserun_key}, CourseUUID: {course_uuid}')
            course_metadata = cls.find_attr(courses_details, 'uuid', course_uuid)
            if not course_metadata:
                logger.info(f'[{cls.LOG_PREFIX}] Metadata not found. CourseUUID: {course_uuid}')
                continue

            course_run = cls.find_attr(course_metadata.get('course_runs'), 'key', courserun_key)
            if not course_run:
                logger.info(
                    f'[{cls.LOG_PREFIX}] Courserun not found. CourserunKey: {courserun_key}, CourseUUID: {course_uuid}'
                )
                continue

            enroll_by = start_date = end_date = None

            # Determine start & end dates for course run
            start_date = course_run.get('start')
            end_date = course_run.get('end')

            # Determine enroll by date for course run.
            enrollment_end = course_run.get('enrollment_end')
            seat = cls.find_best_mode_seat(course_run.get('seats'))
            upgrade_deadline_for_seat = (
                seat.get('upgrade_deadline_override') or seat.get('upgrade_deadline')
                if seat else None
            )
            if upgrade_deadline_for_seat:
                # If upgrade deadline is present, use it as enroll by date.
                enroll_by = upgrade_deadline_for_seat
            elif enrollment_end:
                # If no upgrade deadline, use enrollment end date as enroll by date.
                enroll_by = enrollment_end
            else:
                # If no enroll by date found, log it and continue.
                logger.info(
                    f"[{cls.LOG_PREFIX}] No enroll by date found for course run {courserun_key}"
                )

            # Deduce the course type and product source from course metadata.
            course_key = course_metadata.get('key') or ''
            course_type = course_metadata.get('course_type') or ''
            product_source = course_metadata.get('product_source') or ''
            if product_source:
                product_source = product_source.get('slug')

            # Co-locate courserun metadata into a dict and store it within `courses`.
            course_data = {
                'course_key': course_key,
                'course_type': course_type,
                'product_source': product_source,
                'enroll_by': enroll_by,
                'start_date': start_date,
                'end_date': end_date,
            }
            courses[courserun_key] = course_data

        return courses

    @classmethod
    def store_courses_details(cls, courses_details):
        """
        Store courses metadata in database.
        """
        for courserun_key, course_detail in courses_details.items():
            CourseDetails.objects.update_or_create(
                id=courserun_key,
                defaults=course_detail
            )

    @classmethod
    def find_best_mode_seat(cls, seats):
        """
        Find the seat by best course mode.
        """
        def sort_key(mode):
            """
            Assign a weight to the seat dictionary based on the position of its type in best moode order list.
            """
            mode_type = mode['type']
            if mode_type in BEST_MODE_ORDER:
                return len(BEST_MODE_ORDER) - BEST_MODE_ORDER.index(mode_type)
            else:
                return 0

        sorted_seats = sorted(seats, key=sort_key, reverse=True)
        if not sorted_seats:
            return None
        return sorted_seats[0]

    @classmethod
    def chunks(cls, keys, chunk_size=50):
        """
        Yield chunks of size `chunk_size`.
        """
        for i in range(0, len(keys), chunk_size):
            yield keys[i:i + chunk_size]

    @staticmethod
    def construct_course_key(course_locator):
        """
        Construct course key from course run key.
        """
        return f'{course_locator.org}+{course_locator.course}'

    @classmethod
    def find_attr(cls, iterable, attr_name, attr_value):
        """
        Find value of an attribute from with in an iterable.
        """
        for item in iterable:
            if item[attr_name] == attr_value:
                return item

        return None




from django.db import migrations


def remove_last_refresh_timestamp(apps, schema_editor):
    course_details_import_status = apps.get_model('federated_content_connector', 'CourseDetailsImportStatus')
    course_details_import_status.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('federated_content_connector', '0002_coursedetailsimportstatus'),
    ]

    operations = [
        migrations.RunPython(
            code=remove_last_refresh_timestamp,
            reverse_code=None
        )

    ]
